/* version $Id: foton.cc 6801 2013-03-08 21:43:33Z james.batch@LIGO.ORG $ */
/* -*- mode: c++; c-basic-offset: 4; -*- */
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <TROOT.h>
#include <TApplication.h>
#include <TSystem.h>
#include "foton.hh"
#include "filterwiz/TLGFilterWizard.hh"
#include "goption.hh"

// Allow output of debug info to a file, /tmp/fotonDebug.txt JCB
fstream debugLog ;

   // ROOT object
   TROOT root ("GUI", "Foton");

   using namespace std;
   using namespace filterwiz;


   const char* const help_text =
   "Usage: foton [-w] [-d display] [-p path] [filename]\n"
   "       -w            read-write mode (readonly default)\n"
   "       -l            legacy write mode\n"
   "       -p 'path'     path for filter files\n"
   "       -d 'display'  X windows display\n"
   "       -h            this help\n"
   "\nFoton version " VERSION "\n$Id: foton.cc 6801 2013-03-08 21:43:33Z james.batch@LIGO.ORG $\n";


//////////////////////////////////////////////////////////////////////////
//                                                                      //
// main                                                                 //
//                                                                      //
// Main Program                                                         //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
extern "C"
   int main (int argc, char** argv)
   {
      // JCB - for non-GUI processing of matlab merge file.
      int batch_mode = 0 ;
      int debug_enable = 1 ;

      // JCB - added L for Log file, l for legacy mode.
      gdsbase::option_string opt (argc, argv, "Llvwm:p:d:h");
      // help
      if (opt.help() || (opt.argnum() > 1)) {
         cout << help_text << endl;
         exit (0);
      }
      // This sends deep debugging stuff to a log file. JCB
      if (opt.opt('L'))
      {
	 debugLog.open("/tmp/fotonDebug.txt", fstream::out | fstream::trunc) ;
      }
      // If the file didn't open or the option wasn't selected, send the output to /dev/null.
      if (!debugLog.is_open())
      {
	 debugLog.open("/dev/null") ;
      }

      // silence output if not verbose
      if (!opt.opt ('v')) {
         int i = ::open ("/dev/null", 2);
         (void) dup2 (i, 2); // stderr, cerr  
	 debug_enable = 0 ;
      }

      // filename
      string filename;
      string shortname;
      string filepath;
      if (opt.argnum() == 1) {
         filename = *opt.argbegin();
         if (filename.find ("/") != string::npos) {
            filepath = gSystem->DirName (filename.c_str());
         }
         shortname = gSystem->BaseName (filename.c_str());
      }

      // path
      string path;
      opt.getOpt ('p', path);
      if ((!filepath.empty() && (filepath[0] == '/')) || path.empty()) {
         path = filepath;
      }
      else if (!filepath.empty()) {
         path += "/" + filepath;
      }

      // display
      string display;
      int rargc = 1;
      char* rargv[10];
      rargv[0] = argv[0];
      if (opt.getOpt ('d', display)) {
         rargv[1] = new char [12];
         strcpy (rargv[1], "-display");
         rargv[2] = new char [display.size() + 2];
         strcpy (rargv[2], display.c_str());
         rargc = 3;
      }
      // read-write mode
      bool rw_mode = opt.opt ('w');
      
      // Legacy or aLIGO write mode. - JCB
      bool legacy_mode = opt.opt('l') ;
   
      // Batch Matlab merge option. Requires a matlab file
      // and a filter file, output file is optional and if
      // not specifies the filter file will be used.
      string matlab_file ;
      if (opt.getOpt('m', matlab_file))
      {
         if (matlab_file.empty() || shortname.empty())
         {
            cerr << "foton error: -m requires a " ;
            if (matlab_file.empty())
            {
               cerr << "matlab design file " ;
               if (shortname.empty())
               {
                  cerr << "and a" ;
               }
            }
            if (shortname.empty())
            {
               cerr << "filter file " ;
            }
            cerr << "to be specified on the command line.\n" ;
            return 1 ;
         }
         batch_mode = 1 ;
      }

      // For the matlab batch mode merge, just create a new FilterFile,
      // read the input file, merge the matlab file, and write the results.
      // Return 0 when done if successful.
      if (batch_mode)
      {
	 // realname is the name of the actual input file, in case it's
	 // a symbolic link.
	 string realname ;

	 // Create a new FilterFile object.
	 FilterFile  filter_file ;

	 // Sanity checks. Make sure there's input files to work with.
	 if (filename.empty())
	 {
	    cout << "Error: no filter file specified." << endl ;
	    return -1 ;
	 }
	 if (matlab_file.empty())
	 {
	    cout << "Error: no matlab merge file specified." << endl ;
	    return -1 ;
	 }

	 // Find the real file name in case it's a link. That way the write
	 // function won't overwrite the symbolic link.
	 // If no path was spcified, give getRealFilename a "." to work with.
	 if (filepath.empty())
	    filepath = "." ;
	 realname = filter_file.getRealFilename(filename, filepath) ;
	 if (realname.empty())
	 {
	    cout << "Error: Could not determine path to file " << filename << endl ;
	    return -1 ;
	 }

	 // Set the debug level for the merge functions, this allows errors to
	 // be printed if it fails.
	 filter_file.setMergeDebug(1);

	 // Read the filter file.
	 (void) filter_file.read(realname.c_str()) ;
	 if ((filter_file.errorsEmpty()))
	 {
	    cerr << "Read of " << realname << " succeeded!" << endl ;
	    // Merge the matlab file. (0 return for success)
	    if (filter_file.merge(matlab_file.c_str()) == 0)
	    {
	       cerr << "Merge of " << matlab_file << " successful!" << endl ;
	       // Write the merged filter file.
	       if (!filter_file.write(realname.c_str()))
	       {
		  cerr << "Write failed." << endl ;
		  return -1 ;
	       }
	       else
		  cerr << "Write succeeded!" << endl ;
	    }
	    else
	    {
	       // Error, print the merge errors.
	       cerr << "Merge failed!" << endl ;
	       filter_file.printMergeErrors() ;
	       return -1 ;
	    }
	 }
	 else
	 {
	    // Couldn't read the filter file, print the errors.
	    cerr << "Read Failed!" << endl ;
	    filter_file.printFileErrors() ;
	    return -1 ;
	 }
	 // Successful, so return a 0.
	 return 0 ;
      }

      // silence output if not verbose. cerr done above.
      if (!opt.opt ('v')) {
         int i = ::open ("/dev/null", 2);
         (void) dup2 (i, 1); // stdout, cout
	 debug_enable = 0 ;
      }

      // If this wasn't batch mode, continue with the interactive setup.

      // initializing root environment
      TApplication theApp ("Foton", &rargc, rargv);
   
      // create main window
      bool modal = false;
      std::string fdesign;
      TLGFilterWizard mainWindow (gClient->GetRoot(), modal, "Foton", fdesign,
                           path.empty() ? 0 : path.c_str(), 
                           shortname.empty() ? 0 : shortname.c_str());

      // mainWindow.Setup() always needs to be called regardless
      // of mode.  It causes the filter file to be read (default or 
      // the one specified on the command line).
      mainWindow.Setup() ; // TLGMainWindow::Setup() gets called.

      // Set the merge debug if verbose was specified.
      if (debug_enable)
         mainWindow.setMergeDebug(1) ;

      mainWindow.SetReadOnly (!rw_mode);
      mainWindow.SetLegacyWrite(legacy_mode) ;	// JCB

      // run the GUI
      theApp.Run();
   
      debugLog << "Foton exit." << endl ;
      debugLog.close() ;

      return 0;
   }

